// ParallelCPP.h

#pragma once

using namespace System;                

namespace ParallelCPP 
{
  public enum struct Reduction { Add, Sub, Mul, Div };
  public delegate void CodePart();
  generic <class T> public delegate void Method(T arg);
  generic <class T, class R> public delegate R Function(T arg);

	public ref class ParExec
	{
  public:
    static void For(int iStart, int iEnd, Method<int>^ loopbody)
    {
      #pragma omp parallel for
      for(int i = iStart; i < iEnd; i++)
      {
        loopbody(i);
      }
    }

    static void ForDyn(int iStart, int iEnd, Method<int>^ loopbody)
    {
      #pragma omp parallel for schedule(dynamic)
      for(int i = iStart; i < iEnd; i++)
      {
        loopbody(i);
      }
    }

    static void For(int iStart, int iEnd, int iNumThreads, Method<int>^ loopbody)
    {
      #pragma omp parallel for num_threads(iNumThreads)
      for(int i = iStart; i < iEnd; i++)
      {
        loopbody(i);
      }
    }

    static void ForDyn(int iStart, int iEnd, int iNumThreads, Method<int>^ loopbody)
    {
      #pragma omp parallel for schedule(dynamic) num_threads(iNumThreads)
      for(int i = iStart; i < iEnd; i++)
      {
        loopbody(i);
      }
    }

    static double ForReduceDouble(int iStart, int iEnd, double init, Reduction redu, Function<int, double>^ loopbody)
    {
      double z = init;

      switch(redu)
      {
        case Reduction::Add:
        {
          #pragma omp parallel for reduction(+:z)
          for(int i = iStart; i < iEnd; i++)
          {
            z += loopbody(i);
          }

          break;
        }

        case Reduction::Sub:
        {
          #pragma omp parallel for reduction(+:z)
          for(int i = iStart; i < iEnd; i++)
          {
            z -= loopbody(i);
          }

          break;
        }

        case Reduction::Mul:
        {
          #pragma omp parallel for reduction(+:z)
          for(int i = iStart; i < iEnd; i++)
          {
            z *= loopbody(i);
          }

          break;
        }

        case Reduction::Div:
        {
          #pragma omp parallel for reduction(+:z)
          for(int i = iStart; i < iEnd; i++)
          {
            z /= loopbody(i);
          }

          break;
        }
      }

      return z;
    }

    static void ForReduce(int iStart, int iEnd, double* z, Reduction redu, Function<int, double>^ loopbody)
    {
      double v = *z;

      #pragma omp parallel for reduction(+:v)
      for(int i = iStart; i < iEnd; i++)
      {
        v += loopbody(i);
      }

      *z = v;
    }

    static double ForReduceDouble(int iStart, int iEnd, double init, Reduction redu, int iNumThreads, Function<int, double>^ loopbody)
    {
      double z = init;

      switch(redu)
      {
        case Reduction::Add:
        {
          #pragma omp parallel for reduction(+:z) num_threads(iNumThreads)
          for(int i = iStart; i < iEnd; i++)
          {
            z += loopbody(i);
          }
        }

        case Reduction::Sub:
        {
          #pragma omp parallel for reduction(+:z) num_threads(iNumThreads)
          for(int i = iStart; i < iEnd; i++)
          {
            z -= loopbody(i);
          }
        }

        case Reduction::Mul:
        {
          #pragma omp parallel for reduction(+:z) num_threads(iNumThreads)
          for(int i = iStart; i < iEnd; i++)
          {
            z *= loopbody(i);
          }
        }

        case Reduction::Div:
        {
          #pragma omp parallel for reduction(+:z) num_threads(iNumThreads)
          for(int i = iStart; i < iEnd; i++)
          {
            z /= loopbody(i);
          }
        }
      }

      return z;
    }

    static void Sections(CodePart^ cp1, CodePart^ cp2)
    {
      #pragma omp parallel sections
      {
        #pragma omp section
        {
          cp1();
        }
        #pragma omp section
        {
          cp2();
        }
      }
    }

    static void Sections(CodePart^ cp1, CodePart^ cp2, CodePart^ cp3)
    {
      #pragma omp parallel sections
      {
        #pragma omp section
        {
          cp1();
        }
        #pragma omp section
        {
          cp2();
        }
        #pragma omp section
        {
          cp3();
        }
      }
    }

    static void Sections(CodePart^ cp1, CodePart^ cp2, CodePart^ cp3, CodePart^ cp4)
    {
      #pragma omp parallel sections
      {
        #pragma omp section
        {
          cp1();
        }
        #pragma omp section
        {
          cp2();
        }
        #pragma omp section
        {
          cp3();
        }
        #pragma omp section
        {
          cp4();
        }
      }
    }

    static void Sections(CodePart^ cp1, CodePart^ cp2, int iNumThreads)
    {
      #pragma omp parallel sections num_threads(iNumThreads)
      {
        #pragma omp section
        {
          cp1();
        }
        #pragma omp section
        {
          cp2();
        }
      }
    }

    static void Sections(CodePart^ cp1, CodePart^ cp2, CodePart^ cp3, int iNumThreads)
    {
      #pragma omp parallel sections num_threads(iNumThreads)
      {
        #pragma omp section
        {
          cp1();
        }
        #pragma omp section
        {
          cp2();
        }
        #pragma omp section
        {
          cp3();
        }
      }
    }

    static void Sections(CodePart^ cp1, CodePart^ cp2, CodePart^ cp3, CodePart^ cp4, int iNumThreads)
    {
      #pragma omp parallel sections num_threads(iNumThreads)
      {
        #pragma omp section
        {
          cp1();
        }
        #pragma omp section
        {
          cp2();
        }
        #pragma omp section
        {
          cp3();
        }
        #pragma omp section
        {
          cp4();
        }
      }
    }
  };
}
